@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2011  NTT DATA CORPORATION
  *
- * Version: 2.3.0+   2011/09/29
+ * Version: 2.4.0+   2011/09/29
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License v2 as published by the
@@ -110,9 +110,7 @@ int delete_module(const char *name);
 #define proc_policy_profile          "/sys/kernel/security/tomoyo/profile"
 #define proc_policy_manager          "/sys/kernel/security/tomoyo/manager"
 #define proc_policy_query            "/sys/kernel/security/tomoyo/query"
-#define proc_policy_grant_log        "/sys/kernel/security/tomoyo/grant_log"
-#define proc_policy_reject_log       "/sys/kernel/security/tomoyo/reject_log"
-#define proc_policy_domain_status    "/sys/kernel/security/tomoyo/.domain_status"
+#define proc_policy_audit            "/sys/kernel/security/tomoyo/audit"
 #define proc_policy_process_status   "/sys/kernel/security/tomoyo/.process_status"
 #define proc_policy_self_domain      "/sys/kernel/security/tomoyo/self_domain"
 
@@ -124,18 +122,18 @@ static pid_t pid = 0;
 
 static void clear_status(void)
 {
-	static const char *keywords[] = {
+	static const char * const keywords[] = {
 		"file::execute",
 		"file::open",
 		"file::create",
 		"file::unlink",
+		"file::getattr",
 		"file::mkdir",
 		"file::rmdir",
 		"file::mkfifo",
 		"file::mksock",
 		"file::truncate",
 		"file::symlink",
-		"file::rewrite",
 		"file::mkblock",
 		"file::mkchar",
 		"file::link",
@@ -146,50 +144,8 @@ static void clear_status(void)
 		"file::ioctl",
 		"file::chroot",
 		"file::mount",
-		"file::umount",
+		"file::unmount",
 		"file::pivot_root",
-		"misc::env",
-		"network::inet_udp_bind",
-		"network::inet_udp_connect",
-		"network::inet_tcp_bind",
-		"network::inet_tcp_listen",
-		"network::inet_tcp_connect",
-		"network::inet_tcp_accept",
-		"network::inet_raw_bind",
-		"network::inet_raw_connect",
-		"ipc::signal",
-		"capability::inet_tcp_create",
-		"capability::inet_tcp_listen",
-		"capability::inet_tcp_connect",
-		"capability::use_inet_udp",
-		"capability::use_inet_ip",
-		"capability::use_route",
-		"capability::use_packet",
-		"capability::SYS_MOUNT",
-		"capability::SYS_UMOUNT",
-		"capability::SYS_REBOOT",
-		"capability::SYS_CHROOT",
-		"capability::SYS_KILL",
-		"capability::SYS_VHANGUP",
-		"capability::SYS_TIME",
-		"capability::SYS_NICE",
-		"capability::SYS_SETHOSTNAME",
-		"capability::use_kernel_module",
-		"capability::create_fifo",
-		"capability::create_block_dev",
-		"capability::create_char_dev",
-		"capability::create_unix_socket",
-		"capability::SYS_LINK",
-		"capability::SYS_SYMLINK",
-		"capability::SYS_RENAME",
-		"capability::SYS_UNLINK",
-		"capability::SYS_CHMOD",
-		"capability::SYS_CHOWN",
-		"capability::SYS_IOCTL",
-		"capability::SYS_KEXEC_LOAD",
-		"capability::SYS_PIVOT_ROOT",
-		"capability::SYS_PTRACE",
-		"capability::conceal_mount",
 		NULL
 	};
 	int i;
@@ -200,9 +156,8 @@ static void clear_status(void)
 		exit(1);
 	}
 	for (i = 0; keywords[i]; i++)
-		fprintf(profile_fp,
-			"255-CONFIG::%s={ mode=disabled }\n",
-			keywords[i]);
+		fprintf(profile_fp, "255-CONFIG::%s={ mode=disabled "
+			"grant_log=no reject_log=no }\n", keywords[i]);
 	while (memset(buffer, 0, sizeof(buffer)),
 	       fgets(buffer, sizeof(buffer) - 10, fp)) {
 		const char *mode;
@@ -221,27 +176,16 @@ static void clear_status(void)
 		if (!strcmp(cp, "COMMENT"))
 			mode = "Profile for kernel test\n";
 		else
-			mode = "{ mode=disabled verbose=no }\n";
+			mode = "{ mode=disabled grant_log=no reject_log=no }"
+				"\n";
 		fprintf(profile_fp, "255-%s=%s", cp, mode);
 	}
-	/* fprintf(profile_fp, "255-PREFERENCE::enforcing= penalty=1\n"); */
-	
-	  fprintf(profile_fp, "255-PREFERENCE::learning= verbose=yes\n");
-	  fprintf(profile_fp, "255-PREFERENCE::enforcing= verbose=yes\n");
-	  fprintf(profile_fp, "255-PREFERENCE::permissive= verbose=yes\n");
-	  fprintf(profile_fp, "255-PREFERENCE::disabled= verbose=yes\n");
-	  /*
-	fprintf(profile_fp, "255-PREFERENCE::learning= verbose=no\n");
-	fprintf(profile_fp, "255-PREFERENCE::enforcing= verbose=no\n");
-	fprintf(profile_fp, "255-PREFERENCE::permissive= verbose=no\n");
-	fprintf(profile_fp, "255-PREFERENCE::disabled= verbose=no\n");
-	fprintf(profile_fp, "255-PREFERENCE::learning= max_entry=2048\n");
-	  */
+	fprintf(profile_fp, "255-PREFERENCE={ max_learning_entry=2048 }\n");
 	fflush(profile_fp);
 	fclose(fp);
 }
 
-static void tomoyo_test_init(void)
+static void ccs_test_init(void)
 {
 	pid = getpid();
 	if (access(proc_policy_dir, F_OK)) {
@@ -287,12 +231,12 @@ static void tomoyo_test_init(void)
 	}
 	fprintf(domain_fp, "select pid=%u\n", pid);
 	fprintf(domain_fp, "use_profile 255\n");
-	fprintf(domain_fp, "allow_read/write /sys/kernel/security/tomoyo/domain_policy\n");
-	fprintf(domain_fp, "allow_truncate /sys/kernel/security/tomoyo/domain_policy\n");
-	fprintf(domain_fp, "allow_read/write /sys/kernel/security/tomoyo/exception_policy\n");
-	fprintf(domain_fp, "allow_truncate /sys/kernel/security/tomoyo/exception_policy\n");
-	fprintf(domain_fp, "allow_read/write /sys/kernel/security/tomoyo/profile\n");
-	fprintf(domain_fp, "allow_truncate /sys/kernel/security/tomoyo/profile\n");
+	fprintf(domain_fp, "file read/write/truncate/getattr "
+		"securityfs:/tomoyo/domain_policy\n");
+	fprintf(domain_fp, "file read/write/truncate/getattr "
+		"securityfs:/tomoyo/exception_policy\n");
+	fprintf(domain_fp, "file read/write/truncate/getattr "
+		"securityfs:/tomoyo/profile\n");
 }
 
 static void BUG(const char *fmt, ...)
@@ -311,40 +255,168 @@ static void BUG(const char *fmt, ...)
 		sleep(100);
 }
 
+static char *ccs_freadline(FILE *fp)
+{
+	static char *policy = NULL;
+	int pos = 0;
+	while (1) {
+		static int max_policy_len = 0;
+		const int c = fgetc(fp);
+		if (c == EOF)
+			return NULL;
+		if (pos == max_policy_len) {
+			char *cp;
+			max_policy_len += 4096;
+			cp = realloc(policy, max_policy_len);
+			if (!cp) {
+				BUG("Out of memory");
+				exit(1);
+			}
+			policy = cp;
+		}
+		policy[pos++] = (char) c;
+		if (c == '\n') {
+			policy[--pos] = '\0';
+			break;
+		}
+	}
+	return policy;
+}
+
+static char *ccs_freadline_unpack(FILE *fp)
+{
+	static char *previous_line = NULL;
+	static char *cached_line = NULL;
+	static int pack_start = 0;
+	static int pack_len = 0;
+	if (cached_line)
+		goto unpack;
+	if (!fp)
+		return NULL;
+	{
+		char *pos;
+		unsigned int offset;
+		unsigned int len;
+		char *line = ccs_freadline(fp);
+		if (!line)
+			return NULL;
+		if (sscanf(line, "acl_group %u", &offset) == 1 && offset < 256)
+			pos = strchr(line + 11, ' ');
+		else
+			pos = NULL;
+		if (pos++)
+			offset = pos - line;
+		else
+			offset = 0;
+		if (!strncmp(line + offset, "file ", 5)) {
+			char *cp = line + offset + 5;
+			char *cp2 = strchr(cp + 1, ' ');
+			len = cp2 - cp;
+			if (cp2 && memchr(cp, '/', len)) {
+				pack_start = cp - line;
+				goto prepare;
+			}
+		}
+		return line;
+prepare:
+		pack_len = len;
+		cached_line = strdup(line);
+		if (!cached_line) {
+			BUG("Out of memory");
+			exit(1);
+		}
+	}
+unpack:
+	{
+		char *line = NULL;
+		char *pos = cached_line + pack_start;
+		char *cp = memchr(pos, '/', pack_len);
+		unsigned int len = cp - pos;
+		free(previous_line);
+		previous_line = NULL;
+		if (!cp) {
+			previous_line = cached_line;
+			cached_line = NULL;
+			line = previous_line;
+		} else if (pack_len == 1) {
+			/* Ignore trailing empty word. */
+			free(cached_line);
+			cached_line = NULL;
+		} else {
+			/* Current string is "abc d/e/f ghi". */
+			line = strdup(cached_line);
+			if (!line) {
+				BUG("Out of memory");
+				exit(1);
+			}
+			previous_line = line;
+			/* Overwrite "abc d/e/f ghi" with "abc d ghi". */
+			memmove(line + pack_start + len, pos + pack_len,
+				strlen(pos + pack_len) + 1);
+			/* Overwrite "abc d/e/f ghi" with "abc e/f ghi". */
+			cp++;
+			memmove(pos, cp, strlen(cp) + 1);
+			/* Forget "d/" component. */
+			pack_len -= len + 1;
+			/* Ignore leading and middle empty word. */
+			if (!len)
+				goto unpack;
+		}
+		return line;
+	}
+}
+
 static int write_domain_policy(const char *policy, int is_delete)
 {
-	FILE *fp = fopen(proc_policy_domain_policy, "r");
-	char buffer[8192];
-	int domain_found = 0;
-	int policy_found = 0;
-	memset(buffer, 0, sizeof(buffer));
-	if (!fp) {
-		BUG("Can't read %s", proc_policy_domain_policy);
-		return 0;
+	char *tmp_policy = strdup(policy);
+	if (!tmp_policy) {
+		BUG("Out of memory");
+		exit(1);
 	}
-	if (is_delete)
-		fprintf(domain_fp, "delete ");
-	fprintf(domain_fp, "%s\n", policy);
-	while (fgets(buffer, sizeof(buffer) - 1, fp)) {
-		char *cp = strchr(buffer, '\n');
-		if (cp)
-			*cp = '\0';
-		if (!strncmp(buffer, "<kernel>", 8))
-			domain_found = !strcmp(self_domain, buffer);
-		if (!domain_found)
-			continue;
-		/* printf("<%s>\n", buffer); */
-		if (strcmp(buffer, policy))
-			continue;
-		policy_found = 1;
-		break;
-	}
-	fclose(fp);
-	if (policy_found == is_delete) {
-		BUG("Can't %s %s", is_delete ? "delete" : "append",
-		    policy);
-		return 0;
+	while (1) {
+		FILE *fp = fopen(proc_policy_domain_policy, "r");
+		_Bool domain_found = 0;
+		_Bool policy_found = 0;
+		if (!fp) {
+			BUG("Can't read %s", proc_policy_domain_policy);
+			return 0;
+		}
+		{
+			char *cp = strrchr(tmp_policy, '\t');
+			if (cp)
+				*cp++ = '\0';
+			else
+				cp = tmp_policy;
+			policy = cp;
+		}
+		if (is_delete)
+			fprintf(domain_fp, "delete ");
+		fprintf(domain_fp, "%s\n", policy);
+		while (1) {
+			char *line = ccs_freadline_unpack(fp);
+			if (!line)
+				break;
+			if (!strncmp(line, "<kernel>", 8))
+				domain_found = !strcmp(self_domain, line);
+			if (!domain_found)
+				continue;
+			/* printf("<%s>\n", buffer); */
+			if (strcmp(line, policy))
+				continue;
+			policy_found = 1;
+			while (ccs_freadline_unpack(NULL));
+			break;
+		}
+		fclose(fp);
+		if (policy_found == is_delete) {
+			BUG("Can't %s %s", is_delete ? "delete" : "append",
+			    policy);
+			return 0;
+		}
+		if (policy == tmp_policy)
+			break;
 	}
+	free(tmp_policy);
 	errno = 0;
 	return 1;
 
@@ -352,32 +424,52 @@ static int write_domain_policy(const char *policy, int is_delete)
 
 static int write_exception_policy(const char *policy, int is_delete)
 {
-	FILE *fp = fopen(proc_policy_exception_policy, "r");
-	char buffer[8192];
-	int policy_found = 0;
-	memset(buffer, 0, sizeof(buffer));
-	if (!fp) {
-		BUG("Can't read %s", proc_policy_exception_policy);
-		return 0;
+	char *tmp_policy = strdup(policy);
+	if (!tmp_policy) {
+		BUG("Out of memory");
+		exit(1);
 	}
-	if (is_delete)
-		fprintf(exception_fp, "delete ");
-	fprintf(exception_fp, "%s\n", policy);
-	while (fgets(buffer, sizeof(buffer) - 1, fp)) {
-		char *cp = strchr(buffer, '\n');
-		if (cp)
-			*cp = '\0';
-		if (strcmp(buffer, policy))
-			continue;
-		policy_found = 1;
-		break;
-	}
-	fclose(fp);
-	if (policy_found == is_delete) {
-		BUG("Can't %s %s", is_delete ? "delete" : "append",
-		    policy);
-		return 0;
+	while (1) {
+		FILE *fp = fopen(proc_policy_exception_policy, "r");
+		_Bool policy_found = 0;
+		if (!fp) {
+			BUG("Can't read %s", proc_policy_exception_policy);
+			return 0;
+		}
+		{
+			char *cp = strrchr(tmp_policy, '\t');
+			if (cp)
+				*cp++ = '\0';
+			else
+				cp = tmp_policy;
+			policy = cp;
+		}
+		if (is_delete)
+			fprintf(exception_fp, "delete ");
+		fprintf(exception_fp, "%s\n", policy);
+		while (1) {
+			char *line = ccs_freadline_unpack(fp);
+			if (!line)
+				break;
+			if (!strncmp(line, "<kernel> ", 9))
+				line += 9;
+			/* printf("<%s>\n", buffer); */
+			if (strcmp(line, policy))
+				continue;
+			policy_found = 1;
+			while (ccs_freadline_unpack(NULL));
+			break;
+		}
+		fclose(fp);
+		if (policy_found == is_delete) {
+			BUG("Can't %s %s", is_delete ? "delete" : "append",
+			    policy);
+			return 0;
+		}
+		if (policy == tmp_policy)
+			break;
 	}
+	free(tmp_policy);
 	errno = 0;
 	return 1;
 
@@ -395,12 +487,14 @@ static int set_profile(const int mode, const char *name)
 		BUG("Can't read %s", proc_policy_profile);
 		return 0;
 	}
-	fprintf(profile_fp, "255-CONFIG::%s=%s\n", name, modes[mode]);
+	fprintf(profile_fp, "255-CONFIG::%s={ mode=%s }\n", name, modes[mode]);
 	while (memset(buffer, 0, sizeof(buffer)),
 	       fgets(buffer, sizeof(buffer) - 1, fp)) {
 		char *cp = strchr(buffer, '\n');
 		if (cp)
 			*cp = '\0';
+		if (!strncmp(buffer, "<kernel> ", 9))
+			memmove(buffer, buffer + 9, strlen(buffer + 9) + 1);
 		if (strncmp(buffer, "255-CONFIG::", 12) ||
 		    strncmp(buffer + 12, name, len) ||
 		    buffer[12 + len] != '=')
@@ -411,7 +505,7 @@ static int set_profile(const int mode, const char *name)
 	}
 	fclose(fp);
 	if (!policy_found) {
-		BUG("Can't change profile to 255-CONFIG::%s=%s",
+		BUG("Can't change profile to 255-CONFIG::%s={ mode=%s }",
 		    name, modes[mode]);
 		return 0;
 	}
